/*
 *			D I S P L A Y . C
 *
 *  This section handles the display initialization
 *    and updates for "pmon"
 */
#include <curses.h>
#include <pwd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/file.h>
#ifdef sel
#include <machine/param.h>
#else
#include <sys/param.h>
#endif sel
#include "pmon.h"

#define	CTOK(x)	(ctob(x)/1024)	/* clicks to K bytes */

#define	HOSTLEN	40		/* Length of hostname */

struct	passwd	*pwptr;

/*
 * Return a string describing the given proc p_state
 */
char *
pstat(val)
int val;
{
	static	char	*run2 = "RUN2";

	static char *states[] = {
		"Got Me!",
		"SLEEP",
		"WAIT",
		"RUN",
		"IDL",
		"ZOMB",
		"STOP"
	};

#ifndef	ultrix
#ifdef sel
	if( val == SRUN && pr.p_running == 2 )
		return( run2 );
#else
	if( val == SRUN && pr.p_run == 2 )
		return( run2 );
#endif sel
#endif	ultrix

	if( val > 6 || val < 0 )
		return( states[0] );
	else
		return( states[val] );
}

#ifdef	ultrix
/*
 *	In Ultrix, the flags are in many different places
 *	but bits don't overlap, so we can just OR everything
 *	together.  (See <sys/proc.h>.)
 */
allflags(prp, up)
struct proc *prp;
struct user *up;
{
	unsigned long flags;
#ifndef	SMASTER
	/* 4.0 or later */
	flags = prp->p_vm | prp->p_sched | prp->p_type 
		| prp->p_trace | prp->p_file;
	if (up)
	    flags |= (up->u_oweupc | up->u_sigflag);
#else
	flags = prp->p_flag;
#endif	SIDLEP
	return(flags);
}
/*
 * Return the proc struct flags string.
 */
char *
pflags(flags)
unsigned long flags;
{
	int	i;
	static char buf[100];
	char *bufp;
	static char *states[] = {
		"LOAD",		/* in core */
		"SYS",		/* swapper or pager process */
		"LOCK",		/* process being swapped out */
		"SWAP",		/* save area flag */
		"TRC",		/* process is being traced */
		"WTED",		/* another tracing flag */
		"ULOCK",	/* user settable lock in core */
		"PAGE",		/* process in page wait state */
		"KEEP",		/* another flag to prevent swap out */
		"OMASK",	/* restore old mask after taking signal */
		"WEXIT",	/* working on exiting */
		"PHYSIO",	/* doing physical i/o (bio.c) */
		"VFORK",	/* process resulted from vfork() */
		"VFDONE",	/* another vfork flag */
		"NOVM",		/* no vm, parent in a vfork() */
		"PAGI",		/* init data space on demand, from inode */
		"SEQL",		/* user warned of sequential vm behavior */
		"UANOM",	/* user warned of random vm behavior */
		"TIMO",		/* timing out during sleep */
		"XCTDAT",	/* MIPS: an icache flush on pagein */
		"NOCLDSTP",	/* child stop flag */
		"OWEUPC",	/* owe process an addupc() call at next ast */
		"SEL",		/* selecting; wakeup/waiting danger */
		"LOGIN",	/* a login process (legit child of init) */
		"PTECHG",	/* pte's for process have changed */
		"MASTER",	/* need service by master processor (AMP) */
		"LKDONE",	/* Sys-V file locking applied */
#ifdef	SMASTER
		"EXECDN",	/* exec() done - needed for POSIX pgrp test */
#else
		"FIXADE",	/* fixup unalligned address errors */
#endif
		"EXECDN",	/* exec() done - needed for POSIX pgrp test */
		"IDLEP"		/* idle process */
	};

	buf[0] = NULL;
	for( i = 31; i >= 0; i-- ) {
		if( flags & (1<<i) ) {
			strcat( buf, states[i] );
			strcat( buf, " " );
		}
	}

	return( buf );
}
#else
/*
 * Return the proc struct flags string.
 */
char *
pflags(flags)
unsigned long flags;
{
	int	i;
	static char buf[100];
	char *bufp;
	static char *states[] = {
		"LOAD",		/* in core */
		"SYS",		/* swapper or pager process */
		"LOCK",		/* process being swapped out */
		"SWAP",		/* save area flag */
		"TRC",		/* process is being traced */
		"WTED",		/* another tracing flag */
		"ULOCK",	/* user settable lock in core */
		"PAGE",		/* process in page wait state */
		"KEEP",		/* another flag to prevent swap out */
		"OMASK",	/* restore old mask after taking signal */
		"WEXIT",	/* working on exiting */
		"PHYSIO",	/* doing physical i/o (bio.c) */
		"VFORK",	/* process resulted from vfork() */
		"VFDONE",	/* another vfork flag */
		"NOVM",		/* no vm, parent in a vfork() */
		"PAGI",		/* init data space on demand, from inode */
		"SEQL",		/* user warned of sequential vm behavior */
		"UANOM",	/* user warned of random vm behavior */
		"TIMO",		/* timing out during sleep */
		"Defunct",	/* Was DETACH */
		"OUSIG",	/* using old signal mechanism */
		"OWEUPC",	/* owe process an addupc() call at next ast */
		"SEL",		/* selecting; wakeup/waiting danger */
		"LOGIN",	/* a login process (legit child of init) */
		"PTECHG"	/* pte's for process have changed */
#ifdef sel
		,"JUSTIN",	/* SEL: swapped in and not yet run */
				/* SEL: needed to force pagein before 1st swtch */
		"NOIPU",	/* cannot be run on IPU */
		"EXCSYS",	/* doing excessive system calls */
		"LRIPU",	/* last ran on IPU (try to reuse cache) */
		"IPUCND"	/* is a ipu candidate (due to cpu resched) */
#endif sel
#ifdef vax
		,"none",
		"NODUAL",	/* DUALCPU Do not run this proc on slave cpu */
		"EXCSYS"	/* DUALCPU doing excessive syscalls */
#endif vax
	};

	buf[0] = NULL;
	for( i = 31; i >= 0; i-- ) {
		if( flags & (1<<i) ) {
			strcat( buf, states[i] );
			strcat( buf, " " );
		}
	}

	return( buf );
}
#endif	ultrix

/*
 * Return the kernel csleep() message
 */
char *
pslmsg(loc)
long loc;
{
	static char buf[40];
	char *bufp;

	bufp = buf;
	lseek(kmem, loc, 0);
	do {
		read(kmem, bufp, 1);
	} while(*bufp++);

	return (buf);
}

/*
 *	D I S P I N I T
 *
 * Clears the screen, puts up the info labels
 *  and hostname.
 */
dispinit()
{
	char	hname[HOSTLEN];
	register int i;
	int	ypos;			/* Screen position */

	clear();			/* clear screen */
	gethostname(hname, HOSTLEN);	/* host name in upper left */
	printw(hname);

	/* Put up the labels */
#ifdef TIMERS
	mvprintw( TIMERY, TIMERX, "Timers: Interval, Sec");
#else
	mvprintw( MEMY, MEMX-5, "Mem: text data stack res ( pages )");
#endif TIMERS
	mvprintw( SCHY, SCHX, "pri cpu + nice = upri time slpt");
	mvprintw( SCHY+3, SCHX, "  %%Cpu        User      System");
}

/* "special" process names */
static	char	*pnames[] = {
	"[swapper]",
	"Init",
	"[pagedaemon]"
};

/*
 *	D I S P U P D A T E
 *
 * Updates the dynamic data on the screen.
 */
dispupdate()
{
	int	i;

	/* Print user and command name stuff */
	pwptr = getpwuid( pr.p_uid );
	mvprintw( TOP,LEFT,"User: (%5d) %-10s", pr.p_uid, pwptr->pw_name );
	mvprintw( TOP+1,LEFT,"Cmnd: (%5d) %-15s",
		pr.p_pid, ((pr.p_pid > 2) ? u.u_comm : pnames[ pr.p_pid ]) );
	mvprintw( TOP+2,LEFT,"Pgrp: (%5d) ppid: (%5d)",
		pr.p_pgrp, pr.p_ppid );
	ttyinfo( TOP+3, LEFT );

	/* Display Status */
#if CONTROLT
	if( pr.p_stat == SSLEEP )
		mvprintw( STATY, LEFT, "Stat: %-15s 0x%8x",
			pslmsg(pr.p_slmsg), pr.p_flag );
	else
#endif
#ifdef	ultrix
		mvprintw( STATY, LEFT, "Stat: %-15s 0x%8x",
			pstat(pr.p_stat), allflags(&pr, &u) );
	mvprintw( STATY+1, LEFT,"%-38s", pflags(allflags(&pr, &u)) );
#else
		mvprintw( STATY, LEFT, "Stat: %-15s 0x%8x",
			pstat(pr.p_stat), pr.p_flag );
	mvprintw( STATY+1, LEFT,"%-38s", pflags(pr.p_flag) );
#endif	ultrix

	/* Display Scheduling stuff */
	mvprintw( SCHY+1, SCHX,"%4d %4d %4d %4d %4d %4d",
		pr.p_pri, (unsigned char)pr.p_cpu, pr.p_nice, pr.p_usrpri,
		pr.p_time, pr.p_slptime );
#ifdef	_FIXPT_
	mvprintw( SCHY+4, SCHX, "%9.7f", 100.0 * FIX_TO_DBL(pr.p_pctcpu) );
#else
	mvprintw( SCHY+4, SCHX, "%9.7f", 100.0 * pr.p_pctcpu );
#endif	_FIXPT_
	mvprintw( SCHY+4, SCHX+12, "%3d:%2.2d.%2.2d %3d:%2.2d.%2.2d",
		u.u_ru.ru_utime.tv_sec / 60L,	/* min */
		u.u_ru.ru_utime.tv_sec % 60L,	/* sec */
		u.u_ru.ru_utime.tv_usec / 10000L,	/* hundredths */
		u.u_ru.ru_stime.tv_sec / 60L,
		u.u_ru.ru_stime.tv_sec % 60L,
		u.u_ru.ru_stime.tv_usec / 10000L );

#ifdef TIMERS
	/* Display Timers */
	for( i = 0; i < 3; i++ ) {
		double	intv, value;
		intv = u.u_timer[i].it_interval.tv_sec
			+ u.u_timer[i].it_interval.tv_usec / 1e-6;
		value = u.u_timer[i].it_interval.tv_sec
			+ u.u_timer[i].it_interval.tv_usec / 1e-6;

		mvprintw( TIMERY+i+1, TIMERX, "%10.5f %10.5f\n", intv, value );
	}
#else
	mvprintw( MEMY+1, MEMX, "%4d %4d %4d %4d\n",
		CTOK(pr.p_tsize), CTOK(pr.p_dsize),
		CTOK(pr.p_ssize), CTOK(pr.p_rssize) );
	mvprintw( MEMY+1, MEMX+20, "(%3d/%3d)\n",
		pr.p_rssize,
		pr.p_tsize + pr.p_dsize + pr.p_ssize );
#endif TIMERS

	/* File Descriptors */
	mvprintw( FILEY, LEFT, "Files: %s\n", openfiles() );
	dispfiles( FILEY+1, LEFT );

	/* Signals */
	dispsigs( LEFT, SIGY );

	/* PCB Info */
/*	mvprintw(STATY+9,LEFT,"pc: 0x%7x, usp: 0x%7x, ksp: 0x%7x\n",
		u.u_pcb.pcb_pc, u.u_pcb.pcb_usp, u.u_pcb.pcb_ksp );*/
}

/*
 *	D I S P H E L P
 *
 * The help command.
 */
disphelp()
{
	int	in;

	in = 1 << fileno(stdin);

	clear();			/* clear screen */
	printw("Commands:\n\n");
	printw("^P\tparent process\n");
	printw("^N\tyoungest child\n");
	printw("^F\tyounger sibling\n");
	printw("^B\tolder sibling\n");
	printw("nnn\tgo to process nnn\n");
	printw("f\ttoggle file descriptors\n");
	printw("g\tshow procs in this process group\n");
	printw("t\tshow process tree\n");
	printw("?\tthis message\n");

	printw("\n\n\n");
#ifdef sel
	printw("    d d d d d d   d d d       i s s i s s i\n");
	printw("1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9\n");
	printw("\n");
	printw("H I Q I T I E F K B S S P A T U S T C C T T I X X V P S W\n");
	printw("U N U L R O M P I U E Y I L E R T S O H T T O C F T R T I\n");
	printw("P T I L A T T E L S G S P R R G O T N L I O   P S A O A N\n");
	printw("    T   P       L   V   E M M   P P T D N U   U Z L F C C\n");
	printw("                                                  R   K H\n");
#else
#ifdef	ultrix
	printw("    d d d d d d   d d d       i s s i i s s i         i\n");
	printw("1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9\n");
	printw("\n");
	printw("H I Q I T I E F K B S S P A T U S T C C T T I X X V P W L\n");
	printw("U N U L R O M P I U E Y I L E R T S O H T T O C F T R I O\n");
	printw("P T I L A T T E L S G S P R R G O T N L I O   P S A O N S\n");
	printw("    T   P       L   V   E M M   P P T D N U   U Z L F C T\n");
	printw("                                                  R   H\n");
#else
	printw("    d d d d d d   d d d       i s s i i s s i         i\n");
	printw("1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9\n");
	printw("\n");
	printw("H I Q I T I E F K B S S P A T U S T C C T T I X X V P W\n");
	printw("U N U L R O M P I U E Y I L E R T S O H T T O C F T R I\n");
	printw("P T I L A T T E L S G S P R R G O T N L I O   P S A O N\n");
	printw("    T   P       L   V   E M M   P P T D N U   U Z L F C\n");
	printw("                                                  R   H\n");
#endif	ultrix
#endif sel
	printw("                                                  M\n");
	refresh();
	select( 1, &in, 0, 0, 0 );
	dispinit();
}

/*
 *	L I S T G R O U P
 *
 * Display all processes in same process group.
 */
listgroup()
{
	struct	proc	*saved_procp;
	int	pgrp, in, i;

	saved_procp = procp;

	in = 1 << fileno(stdin);

	clear();			/* clear screen */
	pgrp = pr.p_pgrp;
	printw( "Processes in process group %d\n\n", pgrp );
	procp = (struct proc *)getw(namelist[X_PROC].n_value);
	for (i = 0; i < nproc; i++) {
		lseek(kmem, (long)procp, 0);
		read(kmem, &pr, sizeof(pr));
		/* XXX - is this p_state or p_flag? */
#if	defined(ultrix) && !defined(SMASTER)
		if( allflags(&pr, 0) != 0 && pr.p_pgrp == pgrp) {
#else
		if( pr.p_flag != 0 && pr.p_pgrp == pgrp) {
#endif	ultrix
	        	getupage( &pr );
			printw( "(%5d) %-15s\n",
				pr.p_pid, ((pr.p_pid > 2) ? u.u_comm : pnames[ pr.p_pid ]) );
		}
		procp++;
	}

	/* Restore procp */
	procp = saved_procp;
	refresh();
	select( 1, &in, 0, 0, 0 );
	dispinit();
}

findproc( pid )
int pid;
{
	int	i;
	int oldpid = pr.p_pid;

	procp = (struct proc *)getw(namelist[X_PROC].n_value);
	for (i = 0; i < nproc; i++) {
		lseek(kmem, (long)procp, 0);
		read(kmem, &pr, sizeof(pr));
		/* XXX - is this p_state or p_flag? */
#if	defined(ultrix) && !defined(SMASTER)
		if( allflags(&pr, 0) != 0 && pr.p_pid == pid) {
#else
		if( pr.p_flag != 0 && pr.p_pid == pid) {
#endif	ultrix
			return( 1 );
		}
		procp++;
	}

	/* No such process */
	
	/* MUST RESTORE VALID CONTEXT */
	if (findproc(oldpid) == 0)	/* that process went away */
		findproc(1);	/* init is "always" there */
	return( 0 );
}

show_children( p, level )
struct	proc	*p;
int	level;
{
	struct	proc	*sibling;
	struct	proc	tempproc;
	int	i;

	for( sibling = p; sibling != NULL; sibling = tempproc.p_osptr ) {
		/* read proc and upage info */
		lseek( kmem, (long)sibling, 0 );
		read( kmem, &tempproc, sizeof(tempproc) );
#if	defined(ultrix) && !defined(SMASTER)
		if( allflags(&tempproc, 0) != 0 ) {
#else
		if( tempproc.p_flag != 0 ) {
#endif	ultrix
	        	getupage( &tempproc );
			for( i = 0; i < level; i++ )
				printw("\t");
			printw( "(%5d) %-15s\n",
				tempproc.p_pid, ((tempproc.p_pid > 2) ? u.u_comm : pnames[ tempproc.p_pid ]) );
		}
		/* Display children */
		if( tempproc.p_cptr != NULL )
			show_children( tempproc.p_cptr, level+1 );
		refresh();
	}
}
